-- This module allows the user to display gauges on the HUD.
-- Currently contains only a speedometer.
--
-- For any information or comments on this please go to
-- http://www.beamng.com/threads/4229-Speedometer
--
-- (c) 2013 Grégory Bastogne
-- Licensed under the same terms as Lua (MIT license).
-- http://opensource.org/licenses/MIT


--[[ Module ]]--
local M					= {}

--[[ Locals ]]--
local SU_MPH			= "MPH"
local SU_KMH			= "KMH"
local fonts				= {}
local brushes			= {}
local paintArea			= nil
local canvas			= nil
local maxSpeed			= 160
local splitSpeed		= 10
local fillRatio			= 0.66
local startAngle		= 0
local speedUnit			= SU_MPH
local pOffsetX			= 0.5
local pOffsetY			= 0.5
local radius			= 0.45
local tickOffset		= 0.75
local opacity			= 1
local hide				= false
local hidden			= false
local scale				= 100

--[[ Helper - Converts radians to degrees ]]--
local function radToDeg( rad )
	return math.deg( rad )
	--return 180 * ( rad / math.pi )
end

--[[ Helper - Converts degrees to radians ]]--
local function degToRad( deg )
	return math.rad( deg )
	--return math.pi * ( deg / 180 )
end

--[[ Helper - Returns the positive modulo of a value by a specified mod value ]]--
local function posMod( val, mod )
	return ( ( val % mod ) + mod ) % mod
end

--[[ Helper - Calculates the angle between two points ]]--
local function getAngle( origin, target )
	return posMod( math.atan2( target.y - origin.y, target.x - origin.x ) / ( math.pi / 180 ), 360 )
end

--[[ Helper - Calculates a position ]]--
local function calcPos( center, offset, offAng, size, splitAngle, index )

	--	Calculations
	local pos			= { x = center.x + offset.x, y = center.y + offset.y }
	local ang			= degToRad( posMod( ( splitAngle * index ) + ( splitAngle / 2 ) - 90 + offAng, 360 ) )
	local px			= ( math.cos( ang ) * ( pos.x - center.x ) ) - ( math.sin( ang ) * ( pos.y - center.y ) ) + center.x
	local py			= ( math.sin( ang ) * ( pos.x - center.x ) ) - ( math.cos( ang ) * ( pos.y - center.y ) ) + center.y
	local loc			= { x = -size.width / 2, y = -size.height / 2 }
	local rel			= { x = px + loc.x, y = py + loc.y }
	
	return rel
end

--[[ Helper - Returns the ID of the player's car ]]--
local function getPlayerCarID()
	for objID = 0, BeamEngine:getSlotCount(), 1 do
		local slot	= BeamEngine:getSlot( objID )
		if slot ~= nil and slot.activationMode == 1 then
			return objID
		end
	end
	return 0
end

--[[ Helper - Returns the player's speed in MPH ]]--
local function getPlayerSpeed()

	--	Get the player's car ID
	local carID			= getPlayerCarID()
	if carID == nil then
		return 0, 0
	end
	
	--	Calculate car speed
	local car			= BeamEngine:getSlot( carID )
	local airSpeed		= car:getVelocity():length()
	local kmh			= math.floor( airSpeed * 3.6 )
	local mph			= math.floor( kmh * 0.6214 )
	
	--	Return car speeds
	return kmh, mph
end

--[[ Draw a speed indication ]]--
local function drawSpeed( speed, center, offset, offAng, splitAngle, index )
	
	--	Locals
	local pWidth		= paintArea.width * pOffsetX * 0.15
	local pHeight		= paintArea.height * pOffsetY * 0.15
	local txt			= tostring( speed )
	local len			= string.len( txt )
	local size			= { width = 0, height = 0 }
	local pos			= calcPos( center, offset, offAng, size, splitAngle, index )
	local tPos1			= calcPos( center, { x = offset.x + 10, y = offset.y }, offAng, size, splitAngle, index )
	local tPos2			= calcPos( center, { x = offset.x + pWidth, y = offset.y }, offAng, size, splitAngle, index )
	local stPos1		= calcPos( center, { x = offset.x + 10, y = offset.y }, offAng, size, splitAngle, index + 0.5 )
	local stPos2		= calcPos( center, { x = offset.x + pWidth - 10, y = offset.y }, offAng, size, splitAngle, index + 0.5 )
	
	--	Draw text
	canvas:drawLine( tPos1.x, tPos1.y, tPos2.x, tPos2.y, brushes.gaugeTick )
	canvas:drawLine( stPos1.x, stPos1.y, stPos2.x, stPos2.y, brushes.gaugeTick )
	canvas:drawText( txt, len, pos.x + 1, pos.y + 1, fonts.speedTextShdw )
	canvas:drawText( txt, len, pos.x, pos.y, fonts.speedText )
end

--[[ Draw a needle ]]--
local function drawNeedle( speed, center, offAng, length, splitAngle, index )

	--	Locals
	local pos			= calcPos( center, { x = length, y = 0 }, offAng, { width = 0, height = 0 }, splitAngle, index )
	local o1			= calcPos( center, { x = 0, y = -1 }, offAng, { width = 0, height = 0 }, splitAngle, index )
	local o2			= calcPos( center, { x = 0, y = 1 }, offAng, { width = 0, height = 0 }, splitAngle, index )
	local o3			= calcPos( center, { x = 0, y = -2 }, offAng, { width = 0, height = 0 }, splitAngle, index )
	local o4			= calcPos( center, { x = 0, y = 2 }, offAng, { width = 0, height = 0 }, splitAngle, index )
	local o5			= calcPos( center, { x = 0, y = -3 }, offAng, { width = 0, height = 0 }, splitAngle, index )
	local o6			= calcPos( center, { x = 0, y = 3 }, offAng, { width = 0, height = 0 }, splitAngle, index )
	local o7			= calcPos( center, { x = 0, y = -4 }, offAng, { width = 0, height = 0 }, splitAngle, index )
	local o8			= calcPos( center, { x = 0, y = 4 }, offAng, { width = 0, height = 0 }, splitAngle, index )
	
	--	Draw the needle
	canvas:drawLine( center.x, center.y, pos.x, pos.y, brushes.gaugeNeedle )
	canvas:drawLine( o1.x, o1.y, pos.x, pos.y, brushes.gaugeNeedle )
	canvas:drawLine( o2.x, o2.y, pos.x, pos.y, brushes.gaugeNeedle )
	canvas:drawLine( o3.x, o3.y, pos.x, pos.y, brushes.gaugeNeedle )
	canvas:drawLine( o4.x, o4.y, pos.x, pos.y, brushes.gaugeNeedle )
	canvas:drawLine( o5.x, o6.y, pos.x, pos.y, brushes.gaugeNeedle )
	canvas:drawLine( o5.x, o6.y, pos.x, pos.y, brushes.gaugeNeedle )
	
end

--[[ Draw speed gauge ]]--
local function drawSpeedGauge( dt )
	
	--	Locals
	local pWidth		= paintArea.width
	local pHeight		= paintArea.height
	local pRect			= paintArea.rect
	
	--	Required data
	local kmh, mph		= getPlayerSpeed()
	local speed			= ( speedUnit == SU_MPH ) and mph or kmh
	local speedLabel	= speedUnit
	local speedLblLen	= string.len( speedLabel )
	local speedStr		= tostring( speed )
	local speedLen		= string.len( speedStr )
	local splitAngle	= math.floor( ( 360 / ( maxSpeed / splitSpeed ) ) * fillRatio )
	local center		= { x = ( pWidth * pOffsetX ), y = ( pHeight * pOffsetY ) }
	local offset		= { x = -( pWidth * radius * 0.75 ), y = 0 }
	local offAng		= startAngle + ( 360 - ( 360 * fillRatio ) ) / 2
	
	--	Draw gauge outer
	canvas:drawCircle( center.x, center.y, pWidth * radius, brushes.gaugeOuter )
	
	--	Draw gauge border
	canvas:drawCircle( center.x, center.y, pWidth * radius * 0.95, brushes.gaugeBorder )
	
	--	Draw gauge background
	canvas:drawCircle( center.x, center.y, pWidth * radius * 0.9, brushes.gaugeBkgd )
	
	--	Draw current speed box
	local w				= pWidth * pOffsetX * 0.25
	local h				= 20
	local x				= center.x - w / 2
	local y				= center.y + 10
	local rect			= createSKRect( x, y, w, h )
	local sRect			= createSKRect( x - 2, y - 2, w + 4, h + 4 )
	canvas:drawRect( sRect, brushes.gaugeBorder )
	canvas:drawRect( rect, brushes.gaugeInner )
	
	--	Draw current speed
	canvas:drawText( speedStr, speedLen, center.x + 1, center.y + 25 + 1, fonts.speedTextShdw )
	canvas:drawText( speedStr, speedLen, center.x, center.y + 25, fonts.speedText )
	
	--	Draw needle circle
	canvas:drawCircle( center.x, center.y, pWidth * pOffsetX * 0.05, brushes.gaugeBorder )
	
	--	Draw speed label text
	canvas:drawText( speedLabel, speedLblLen, center.x + 1, center.y - 20 + 1, fonts.gaugeTextShdw )
	canvas:drawText( speedLabel, speedLblLen, center.x, center.y - 20, fonts.gaugeText )
	
	--	Draw needle
	local index			= speed / splitSpeed
	drawNeedle( speed, center, offAng, offset.x, splitAngle, index )
	
	--	Draw speed indications with ticks
	for speedTick = 0, maxSpeed, splitSpeed do
		local index		= speedTick / splitSpeed
		drawSpeed( speedTick, center, { x = -( pWidth * radius * tickOffset ), y = 0  }, offAng, splitAngle, index )
	end
end

--[[ Initialize fonts & brushes ]]--
local function init()

	--	Gauge border brush
	local brush			= SkPaint()
	brush:setAntiAlias( true )
	brush:setARGB( math.ceil( opacity * 200 ), 0, 0, 0 )
	brushes.gaugeBorder	= brush

	--	Gauge outer brush
	local brush			= SkPaint()
	brush:setAntiAlias( true )
	brush:setARGB( math.ceil( opacity * 200 ), 236, 237, 237 )
	brushes.gaugeOuter	= brush
	
	--	Gauge inner brush
	local brush			= SkPaint()
	brush:setAntiAlias( true )
	brush:setARGB( math.ceil( opacity * 255 ), 236, 235, 239 )
	brushes.gaugeInner	= brush
	
	--	Gauge background brush
	local brush			= SkPaint()
	brush:setAntiAlias( true )
	brush:setARGB( math.ceil( opacity * 255 ), 252, 222, 84 )
	brushes.gaugeBkgd	= brush
	
	--	Gauge needle brush
	local brush			= SkPaint()
	brush:setAntiAlias( true )
	brush:setARGB( math.ceil( opacity * 255 ), 150, 49, 0 )
	brushes.gaugeNeedle	= brush
	
	--	Gauge tick brush
	local brush			= SkPaint()
	brush:setAntiAlias( true )
	brush:setARGB( math.ceil( opacity * 255 ), 0, 0, 0 )
	brushes.gaugeTick	= brush
	
	--	Gauge text font
	local typeFace		= SkTypeface.CreateFromFile( "Roboto-Light.ttf", kBoldItalic )
	local font			= SkPaint()
	font:setAntiAlias( true )
	font:setColor( SK_ColorBLACK )
	font:setTextAlign( kCenter_Align )
	font:setTextSize( 20 )
	font:setTypeface( typeFace )
	fonts.gaugeText		= font
	
	--	Gauge text shadow
	local font			= SkPaint()
	font:setAntiAlias( true )
	font:setColor( SK_ColorWHITE )
	font:setTextAlign( kCenter_Align )
	font:setTextSize( 20 )
	font:setTypeface( typeFace )
	fonts.gaugeTextShdw	= font
	
	--	Speed text font
	local font			= SkPaint()
	font:setAntiAlias( true )
	font:setColor( SK_ColorBLACK )
	font:setTextAlign( kCenter_Align )
	font:setTextSize( 13 )
	font:setTypeface( typeFace )
	fonts.speedText		= font
	
	--	Speed text shadow
	local font			= SkPaint()
	font:setAntiAlias( true )
	font:setColor( SK_ColorWHITE )
	font:setTextAlign( kCenter_Align )
	font:setTextSize( 13 )
	font:setTypeface( typeFace )
	fonts.speedTextShdw	= font

end

--[[ Update ]]--
local function update( dt )

	--	Get paint area
	if paintArea == nil then
		paintArea		= BeamEngine.canvasTexture
	end
	
	--	Paint area is required
	if not paintArea then return end
	
	--	Get the canvas
	canvas				= paintArea:getCanvas()
	
	--	Clear area
	paintArea:clear()
	
	--	Do not draw if hidden
	if hidden and hide then return end
	
	--	Do the drawings
	drawSpeedGauge( dt )
	
	--	Hide if required
	if hide and scale > 0 then
		canvas:scale( 0.8, 0.8 )
		scale			= math.max( 0, scale * 0.8 )
		if scale <= 0.1 then hidden = true end
		
	--	Show if required
	elseif not hide and scale < 100 then
		hidden			= false
		canvas:scale( 1.2, 1.2 )
		scale			= math.min( 100, scale * 1.2 )
	end

end

--[[ GUI callback ]]--
local function GUICallback( mode, str )

	--	Extract args
	local args			= unserialize(str)
	
	--	Reset
	if mode == "button" and args.button == "btReset" then
		maxSpeed		= 160
		splitSpeed		= 10
		fillRatio		= 0.66
		startAngle		= 0
		speedUnit		= SU_MPH
		pOffsetX		= 0.5
		pOffsetY		= 0.5
		radius			= 0.45
		tickOffset		= 0.75
		opacity			= 1
		init()
		M.showGUI()
	end
	
	--	Show / hide
	if mode == "button" and args.button == "btShowHide" then
		if not hide and not hidden then hide = true
		elseif hide and hidden then hide = false end
		M.showGUI()
		return
	end
	
	--	Close button pressed
	if args and args.done == 1 then return end
	
	--	Chose item in dropbox or closed
	if mode == "apply" then
		maxSpeed		= tonumber( args.maxSpeed )
		splitSpeed		= tonumber( args.splitSpeed )
		fillRatio		= tonumber( args.fillRatio )
		startAngle		= tonumber( args.startAngle )
		speedUnit		= tostring( args.speedUnit )
		pOffsetX		= tonumber( args.offsetX )
		pOffsetY		= tonumber( args.offsetY )
		radius			= tonumber( args.radius )
		tickOffset		= tonumber( args.tickOffset )
		
		--	Update opacity
		local oldOp		= opacity
		opacity			= tonumber( args.opacityValue )
		if opacity ~= oldOp then
			init()
		end
	end
	
end

--[[ Show GUI ]]--
local function showGUI()

	--	GUI definition
	local g				= [[beamngguiconfig 1
callback system gauges.GUICallback
title Speedometer Configuration
 
container
	type = verticalStack
	name = root
	
control
	type = chooser
	name = speedUnit
	description = Speed unit
	selection = ]] .. speedUnit .. [[

	option = ]] .. SU_MPH .. [[ ]] .. SU_MPH .. [[ 
	option = ]] .. SU_KMH .. [[ ]] .. SU_KMH .. [[ 
	
control
	type = chooser
	name = maxSpeed
	description = Max speed
	selection = ]] .. maxSpeed .. [[

	option = 120 120
	option = 140 140
	option = 160 160
	option = 170 170
	option = 180 180
	option = 190 190
	option = 200 200
	option = 210 210
	option = 220 220
	option = 250 250
	option = 300 300
  
control
	type = chooser
	name = splitSpeed
	description = Split speed
	selection = ]] .. splitSpeed .. [[

	option = 5 5
	option = 10 10
	option = 20 20
	option = 30 30
	option = 40 40
	option = 50 50
	option = 60 60
	option = 70 70
	option = 80 80
	option = 90 90
	option = 100 100
  
control
	type = chooser
	name = fillRatio
	description = Fill ratio
	selection = ]] .. fillRatio .. [[

	option = 0.25 25%
	option = 0.3 30%
	option = 0.33 33%
	option = 0.5 50%
	option = 0.66 66%
	option = 0.7 70%
	option = 0.75 75%
	option = 1 100%
	
control
	type = chooser
	name = startAngle
	description = Start angle
	selection = ]] .. startAngle .. [[

	option = 0 0
	option = 15 15
	option = 30 30
	option = 45 45
	option = 60 60
	option = 75 75
	option = 90 90
	option = 105 105
	option = 120 120
	option = 135 135
	option = 150 150
	option = 165 165
	option = 180 180
	option = 195 195
	option = 210 210
	option = 225 225
	option = 240 240
	option = 255 255
	option = 270 270
	option = 285 258
	option = 300 300
	option = 315 315
	option = 330 330
	option = 345 345
	
control
	type = chooser
	name = offsetX
	description = Horizontal offset
	selection = ]] .. pOffsetX .. [[

	option = 0.25 25%
	option = 0.5 50%
	option = 0.75 75%
	
control
	type = chooser
	name = offsetY
	description = Vertical offset
	selection = ]] .. pOffsetY .. [[

	option = 0.25 25%
	option = 0.5 50%
	option = 0.75 75%
	
control
	type = chooser
	name = radius
	description = Radius
	selection = ]] .. radius .. [[

	option = 0.1 10%
	option = 0.2 20%
	option = 0.25 25%
	option = 0.33 33%
	option = 0.45 45%
	option = 0.5 50%
	
control
	type = chooser
	name = tickOffset
	description = Tick offset
	selection = ]] .. tickOffset .. [[

	option = 0.25 25%
	option = 0.33 33%
	option = 0.5 50%
	option = 0.66 66%
	option = 0.75 75%
	option = 0.90 90%
	option = 1.1 110%
	
control
	type = chooser
	name = opacityValue
	description = Opacity
	selection = ]] .. opacity .. [[

	option = 0 0%
	option = 0.1 10%
	option = 0.25 25%
	option = 0.33 33%
	option = 0.5 50%
	option = 0.66 66%
	option = 0.75 75%
	option = 0.90 90%
	option = 1 100%

control
	type = button
	name = btShowHide
	icon = tools/gui/images/]] .. ( hide and "visible_i.png" or "visible_d.png" ) .. [[ 
	description = ]] .. ( hide and "Show" or "Hide" ) .. [[ 
	
control
	type = button
	name = btReset
	icon = tools/gui/images/reset-icon_d.png
	description = Reset
	
control
	type = doneButton
	name = btClose
	icon = tools/gui/images/iconCancel.png
	description = Close
               
]]
	gameEngine:showGUI( g )
end

--[[ Public interface --]]
M.init					= init
M.update				= update
M.showGUI				= showGUI
M.GUICallback			= GUICallback

--[[ Return module ]]--
return M